-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}
M.type = "auxilliary"
M.relevantDevice = "mainEngine"

local wheels = {} --cache for holding all our wheels
local otherWheelOnAxle = {}

local frontLeftWheelName = ""
local frontRightWheelName = ""
local rearRightWheelName = ""
local rearLeftWheelName = ""


local mainEngine = nil

local tcsHandling = {
  driveWheel = 0, --0 FWD 1 RWD
  speedDiffLimit = 0, --speed diff before throttle reduction
  speedRatioLimit = 1, --speed ratio before throttle reduction
  speedMargin = 0, --above limit this amount max throttle reduction
  minEngineRPM = 0, --deactivate TCS when engine RPM lower than this
  throttleCut = 0, --TCS cut throttle max amount
  brakeSwitch = 0, --deactivate TCS when brake higher than this
}

local function updateGFX()

  
  local wheelSpeeds = {}


  local frontAxleAvgSpeed = 0
  local rearAxleAvgSpeed = 0
  local driveAxleSpeed = 0
  local nonAxleSpeed = 0
  local driveDirection = 0 --1 forward -1 backward
  local axleSpeedRatio = 0
  local axleSpeedDiff = 0
  local tcsDiffRate = 0
  local tcsRatioRate = 0
  local tcsRate = 0
  
  for _,w in pairs(powertrain.wheels) do
	local wheelSpeed = w.speedSmoother:get(w.angularVelocity * w.wheelDir)
	wheelSpeeds[w.name] = wheelSpeed
	--log('I', "Wheel Name ", w.name)
  end
  
  frontAxleAvgSpeed = (wheelSpeeds[frontLeftWheelName] + wheelSpeeds[frontRightWheelName]) * 0.5
  rearAxleAvgSpeed = (wheelSpeeds[rearLeftWheelName] + wheelSpeeds[rearRightWheelName]) * 0.5

  driveAxleSpeed = frontAxleAvgSpeed * (1 - tcsHandling.driveWheel) + rearAxleAvgSpeed * tcsHandling.driveWheel
  nonAxleSpeed = frontAxleAvgSpeed * tcsHandling.driveWheel + rearAxleAvgSpeed * (1 - tcsHandling.driveWheel)

  driveDirection = clamp(driveAxleSpeed,-1,1)

  axleSpeedDiff = math.max(driveAxleSpeed * driveDirection,0) - math.max(nonAxleSpeed * driveDirection,0)
  axleSpeedRatio = math.max(driveAxleSpeed * driveDirection,0) / (math.max(nonAxleSpeed * driveDirection,0) + 1e-30)
  --log('I', "axleSpeedDiff ", axleSpeedDiff)
  --log('I', "speedDiffLimit ", tcsHandling.speedDiffLimit)
  --log('I', "axleSpeedRatio ", axleSpeedRatio)
  --log('I', "speedRatioLimit ", tcsHandling.speedRatioLimit)
  --log('I', "speedMargin ", tcsHandling.speedMargin)

  tcsDiffRate = clamp((axleSpeedDiff - tcsHandling.speedDiffLimit) / (tcsHandling.speedDiffLimit * tcsHandling.speedMargin),0,1)
  tcsRatioRate = clamp((axleSpeedRatio - tcsHandling.speedRatioLimit) / ((tcsHandling.speedRatioLimit - 1) * tcsHandling.speedMargin),0,1)
  tcsRate = math.min(tcsDiffRate,tcsRatioRate) 
  --log('I', "tcsRate ", tcsRate)

  if input.parkingbrake < tcsHandling.brakeSwitch and input.brake < tcsHandling.brakeSwitch and engine.outputAV1 > tcsHandling.minEngineRPM * 0.104719755 then --activate TCS
    input.throttle = input.throttle * (1 - tcsHandling.throttleCut * tcsRate)
  end  
    
  --log('I', "driveWheel ", tcsHandling.driveWheel)

end


local function init(jbeamData)

  engine = powertrain.getDevice(jbeamData.engineName or "mainEngine")

  if engine then
    tcsHandling.driveWheel = jbeamData.tcsDriveWheel or 1 --0 is FWD 1 is RWD
    tcsHandling.speedDiffLimit = jbeamData.tcsSpeedDiffLimit or 5 --speed diff before throttle reduction
    tcsHandling.speedRatioLimit = jbeamData.tcsSpeedRatioLimit or 1.05 --speed ratio before throttle reduction
    tcsHandling.speedMargin = jbeamData.tcsSpeedMargin or 0.5 --above limit this amount max throttle reduction
    tcsHandling.minEngineRPM = jbeamData.tcsMinEngineRPM or 2500 --deactivate TCS when engine RPM lower than this
    tcsHandling.throttleCut = jbeamData.tcsThrottleCut or 0.9 --TCS reduces throttle amount
    tcsHandling.brakeSwitch = jbeamData.tcsBrakeSwitch or 0.2 --deactivate TCS when brake higher than this

    M.updateGFX = updateGFX
  end

  local avgWheelPos = vec3(0,0,0)
  local wheelCount = 0
  wheels = {}
  
  for _,wd in pairs(powertrain.wheels) do
    local name = wd.name
    wheels[name] = wd
    wheels[name].speedSmoother = newExponentialSmoothing(50)
    --log('I', "Wheel Name ", name)
  end
  
  
  for _,w in pairs(powertrain.wheels) do
    
    local wheelNodePos = v.data.nodes[wheels[w.name].node1].pos --find the wheel position
    avgWheelPos = avgWheelPos + wheelNodePos --sum up all positions
    wheelCount = wheelCount + 1
  end
  
  local vectorForward = vec3(v.data.nodes[v.data.refNodes[0].ref].pos) - vec3(v.data.nodes[v.data.refNodes[0].back].pos) -- vec3(obj:getDirectionVector()) --vector facing forward
  local vectorUp = vec3(v.data.nodes[v.data.refNodes[0].up].pos) - vec3(v.data.nodes[v.data.refNodes[0].ref].pos)

  local vectorRight = vectorForward:cross(vectorUp) --vector facing to the right

  --iterate over all wheels 
  for _,w in pairs(powertrain.wheels) do
    local wheelNodePos = vec3(v.data.nodes[wheels[w.name].node1].pos) --find the wheel position
    local wheelVector = wheelNodePos - avgWheelPos --create a vector from our "center" to the wheel
    local dotForward = vectorForward:dot(wheelVector) --calculate dot product of said vector and forward vector
    local dotLeft = vectorRight:dot(wheelVector) --calculate dot product of said vector and left vector

    if dotForward >= 0 then
      if dotLeft >= 0 then
        
		frontRightWheelName = w.name --this case can only mean it's our front right wheel
      else
        frontLeftWheelName = w.name -- ...
      end
    else
      if dotLeft >= 0 then
        rearRightWheelName = w.name-- ...
      else
        rearLeftWheelName = w.name-- ...
      end
    end
  end


end

M.init = init
--M.reset = nop
M.updateGFX = updateGFX

return M